"""InstanceFile and InstanceFileManager."""
import os
import re
import itertools
from pathlib import Path
from schema_enforcer.utils import find_files, load_file

SCHEMA_TAG = "jsonschema"


class InstanceFileManager:  # pylint: disable=too-few-public-methods
    """InstanceFileManager."""

    def __init__(self, config):
        """Initialize the interface File manager.

        The file manager will locate all potential instance files in the search directories.

        Args:
            config (pydantic.BaseSettings): The Pydantec settings object.
        """
        self.instances = []
        self.config = config

        # Find all instance files
        # TODO need to load file extensions from the config
        instance_files = find_files(
            file_extensions=config.data_file_extensions,
            search_directories=config.data_file_search_directories,
            excluded_filenames=config.data_file_exclude_filenames,
            excluded_directories=[config.main_directory],
            return_dir=True,
        )

        # For each instance file, check if there is a static mapping defined in the config
        # Create the InstanceFile object and save it
        for root, filename in instance_files:
            matches = set()
            if filename in config.schema_mapping:
                matches.update(config.schema_mapping[filename])

            instance = InstanceFile(root=root, filename=filename, matches=matches)
            self.instances.append(instance)

    def add_matches_by_property_automap(self, schema_manager):
        """Adds schema_ids to matches by automapping top level schema properties to top level keys in instance data.

        Args:
            schema_manager (schema_enforcer.schemas.manager.SchemaManager): Schema manager oject
        """
        for instance in self.instances:
            instance.add_matches_by_property_automap(schema_manager)

    def print_schema_mapping(self):
        """Print in CLI the matches for all instance files."""
        print("{:50} Schema ID".format("Structured Data File"))  # pylint: disable=consider-using-f-string
        print("-" * 80)
        print_strings = []
        for instance in self.instances:
            filepath = f"{instance.path}/{instance.filename}"
            print_strings.append(f"{filepath:50} {sorted(instance.matches)}")
        print("\n".join(sorted(print_strings)))


class InstanceFile:
    """Class to manage an instance file."""

    def __init__(self, root, filename, matches=None):
        """Initializes InstanceFile object.

        Args:
            root (string): Absolute path to the directory where the schema file is located.
            filename (string): Name of the file.
            matches (set, optional): Set of schema IDs that matches with this Instance file. Defaults to None.
        """
        self.data = None
        self.path = root
        self.full_path = os.path.realpath(root)
        self.filename = filename

        # Internal vars for caching data
        self._top_level_properties = set()

        if matches:
            self.matches = matches
        else:
            self.matches = set()

        self._add_matches_by_decorator()

    @property
    def top_level_properties(self):
        """Return a list of top level properties in the structured data defined by the data pulled from _get_content.

        Returns:
            set: Set of the strings of top level properties defined by the data file
        """
        if not self._top_level_properties:
            content = self._get_content()
            self._top_level_properties = set(content.keys())

        return self._top_level_properties

    def _add_matches_by_decorator(self, content=None):
        """Add matches which declare schema IDs they should adhere to using a decorator comment.

        If a line of the form # jsonschema: <schema_id>,<schema_id> is defined in the data file, the
        schema IDs will be added to the list of schema IDs the data will be checked for adherence to.

        Args:
            content (string, optional): Content of the file to analyze. Default to None.

        Returns:
            set(string): Set of matches (strings of schema_ids) found in the file.
        """
        if not content:
            content = self._get_content(structured=False)

        matches = set()

        if SCHEMA_TAG in content:
            line_regexp = r"^#.*{0}:\s*(.*)$".format(SCHEMA_TAG)  # pylint: disable=consider-using-f-string
            match = re.match(line_regexp, content, re.MULTILINE)
            if match:
                matches = {x.strip() for x in match.group(1).split(",")}

        self.matches.update(matches)

    def _get_content(self, structured=True):
        """Returns the content of the instance file.

        Args:
            structured (bool): Return structured data if true. If false returns the string representation of the data
            stored in the instance file. Defaults to True.

        Returns:
            dict, list, or str: File Contents. Dict or list if structured is set to True. Otherwise returns a string.
        """
        file_location = os.path.join(self.full_path, self.filename)

        if not structured:
            return Path(file_location).read_text(encoding="utf-8")

        return load_file(file_location)

    def add_matches_by_property_automap(self, schema_manager):
        """Adds schema_ids to self.matches by automapping top level schema properties to top level keys in instance data.

        Args:
            schema_manager (schema_enforcer.schemas.manager.SchemaManager): Schema manager oject
        """
        matches = set()

        for schema_id, schema_obj in schema_manager.iter_schemas():
            if schema_obj.top_level_properties.intersection(self.top_level_properties):
                matches.add(schema_id)

        self.matches.update(matches)

    def validate(self, schema_manager, strict=False):
        """Validate this instance file with all matching schema in the schema manager.

        Args:
            schema_manager (SchemaManager): A SchemaManager object.
            strict (bool, optional): True is the validation should automatically flag unsupported element. Defaults to False.

        Returns:
            iterator: Iterator of ValidationErrors returned by schema.validate.
        """
        # TODO need to add something to check if a schema is missing
        # Create new iterator chain to be able to aggregate multiple iterators
        errs = itertools.chain()

        # Go over all schemas and skip any schema not present in the matches
        for schema_id, schema in schema_manager.iter_schemas():
            if schema_id not in self.matches:
                continue
            schema.validate(self._get_content(), strict)
            results = schema.get_results()
            errs = itertools.chain(errs, results)
            schema.clear_results()

        return errs



"""main cli commands."""
import sys

import click
from termcolor import colored

from schema_enforcer.utils import MutuallyExclusiveOption
from schema_enforcer import config
from schema_enforcer.schemas.manager import SchemaManager
from schema_enforcer.instances.file import InstanceFileManager
from schema_enforcer.utils import error
from schema_enforcer.exceptions import InvalidJSONSchema


@click.group()
def main():
    """SCHEMA ENFORCER.

    This tool is used to ensure data adheres to a schema definition. The data can come
    from YAML files, JSON files, or an Ansible inventory. The schema to which the data
    should adhere can currently be defined using the JSONSchema language in YAML or JSON
    format.
    """


@click.option("--show-pass", default=False, help="Shows validation checks that passed", is_flag=True, show_default=True)
@click.option(
    "--strict",
    default=False,
    help="Forces a stricter schema check that warns about unexpected additional properties",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--show-checks",
    default=False,
    help="Shows the schemas to be checked for each structured data file",
    is_flag=True,
    show_default=True,
)
@main.command()
def validate(show_pass, show_checks, strict):  # noqa D205
    """Validates instance files against defined schema.

    \f

    Args:
        show_pass (bool): show successful schema validations
        show_checks (bool): show schemas which will be validated against each instance file
        strict (bool): Forces a stricter schema check that warns about unexpected additional properties
    """
    config.load()

    # ---------------------------------------------------------------------
    # Load Schema(s) from disk
    # ---------------------------------------------------------------------
    try:
        smgr = SchemaManager(config=config.SETTINGS)
    except InvalidJSONSchema as exc:
        error(str(exc))
        sys.exit(1)

    if not smgr.schemas:
        error("No schemas were loaded")
        sys.exit(1)

    # ---------------------------------------------------------------------
    # Load Instances
    # ---------------------------------------------------------------------
    ifm = InstanceFileManager(config=config.SETTINGS)

    if not ifm.instances:
        error("No instance files were found to validate")
        sys.exit(1)

    if config.SETTINGS.data_file_automap:
        ifm.add_matches_by_property_automap(smgr)

    if show_checks:
        ifm.print_schema_mapping()
        sys.exit(0)

    error_exists = False
    for instance in ifm.instances:
        for result in instance.validate(smgr, strict):

            result.instance_type = "FILE"
            result.instance_name = instance.filename
            result.instance_location = instance.path

            if not result.passed():
                error_exists = True
                result.print()

            elif result.passed() and show_pass:
                result.print()

    if not error_exists:
        print(colored("ALL SCHEMA VALIDATION CHECKS PASSED", "green"))
    else:
        sys.exit(1)


@click.option(
    "--list",
    "list_schemas",
    default=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["generate_invalid", "check", "schema-id", "dump"],
    help="List all available schemas",
    is_flag=True,
)
@click.option(
    "--dump",
    "dump_schemas",
    default=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["generate_invalid", "check", "list"],
    help="Dump full schema for all schemas or schema-id",
    is_flag=True,
)
@click.option(
    "--check",
    default=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["generate_invalid", "list", "dump"],
    help="Validates that all schemas are valid (spec and unit tests)",
    is_flag=True,
)
@click.option(
    "--generate-invalid",
    default=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["check", "list", "dump"],
    help="Generates expected invalid result from a given schema [--schema-id] and data defined in a data file",
    is_flag=True,
)
@click.option(
    "--schema-id", default=None, cls=MutuallyExclusiveOption, mutually_exclusive=["list"], help="The name of a schema."
)
@main.command()
def schema(check, generate_invalid, list_schemas, schema_id, dump_schemas):  # noqa: D417,D301,D205
    """Manage your schemas.

    \f

    Args:
        check (bool): Validates that all schemas are valid (spec and unit tests)
        generate_invalid (bool): Generates expected invalid data from a given schema
        list_schemas (bool): List all available schemas
        schema_id (str): Name of schema to evaluate
        dump_schemas (bool): Dump all schema data or a single schema if schema_id is provided
    """
    if not check and not generate_invalid and not list_schemas and not schema_id and not dump_schemas:
        error(
            "The 'schema' command requires one or more arguments. You can run the command 'schema-enforcer schema --help' to see the arguments available."
        )
        sys.exit(1)

    config.load()

    # ---------------------------------------------------------------------
    # Load Schema(s) from disk
    # ---------------------------------------------------------------------
    try:
        smgr = SchemaManager(config=config.SETTINGS)
    except InvalidJSONSchema as exc:
        error(str(exc))
        sys.exit(1)

    if not smgr.schemas:
        error("No schemas were loaded")
        sys.exit(1)

    if list_schemas:
        smgr.print_schemas_list()
        sys.exit(0)

    if dump_schemas:
        smgr.dump_schema(schema_id)
        sys.exit(0)

    if generate_invalid:
        if not schema_id:
            sys.exit("Please indicate the schema you'd like to generate invalid data for using the --schema-id flag")
        smgr.generate_invalid_tests_expected(schema_id=schema_id)
        sys.exit(0)

    if check:
        smgr.test_schemas()
        sys.exit(0)


@main.command()
@click.option("--inventory", "-i", help="Ansible inventory file.", required=False)
@click.option("--host", "-h", "limit", help="Limit the execution to a single host.", required=False)
@click.option("--show-pass", default=False, help="Shows validation checks that passed", is_flag=True, show_default=True)
@click.option(
    "--show-checks",
    default=False,
    help="Shows the schemas to be checked for each ansible host",
    is_flag=True,
    show_default=True,
)
def ansible(
    inventory, limit, show_pass, show_checks
):  # pylint: disable=too-many-branches,too-many-locals,too-many-locals,too-many-statements  # noqa: D417,D301
    """Validate the hostvars for all hosts within an Ansible inventory.

    The hostvars are dynamically rendered based on groups to which each host belongs.
    For each host, if a variable `schema_enforcer_schema_ids` is defined, it will be used
    to determine which schemas should be use to validate each key. If this variable is
    not defined, the hostvars top level keys will be automatically mapped to a schema
    definition's top level properties to automatically infer which schema should be used
    to validate which hostvar.
    \f

    Args:
        inventory (string): The name of the file used to construct an ansible inventory.
        limit (string, None): Name of a host to limit the execution to.
        show_pass (bool): Shows validation checks that pass. Defaults to False.
        show_checks (bool): Shows the schema ids each host will be evaluated against.

    Example:
        $ cd examples/ansible
        $ ls -ls
        total 8
        drwxr-xr-x  5 damien  staff   160B Jul 25 16:37 group_vars
        drwxr-xr-x  4 damien  staff   128B Jul 25 16:37 host_vars
        -rw-r--r--  1 damien  staff    69B Jul 25 16:37 inventory.ini
        drwxr-xr-x  4 damien  staff   128B Jul 25 16:37 schema
        $ schema-enforcer ansible -i inventory.ini
        Found 4 hosts in the inventory
        FAIL | [ERROR] False is not of type 'string' [HOST] spine1 [PROPERTY] dns_servers:0:address
        FAIL | [ERROR] False is not of type 'string' [HOST] spine2 [PROPERTY] dns_servers:0:address
        $ schema-enforcer ansible -i inventory.ini -h leaf1
        Found 4 hosts in the inventory
        ALL SCHEMA VALIDATION CHECKS PASSED
        $ schema-enforcer ansible -i inventory.ini -h spine1 --show-pass
        Found 4 hosts in the inventory
        FAIL | [ERROR] False is not of type 'string' [HOST] spine1 [PROPERTY] dns_servers:0:address
        PASS | [HOST] spine1 [SCHEMA ID] schemas/interfaces
    """
    # Ansible is currently always installed by schema-enforcer. This was added in the interest of making ansible an
    # optional dependency. We decided to make two separate packages installable via PyPi, one with ansible, one without.
    # This has been left in the code until such a time as we implement the change to two packages so code will not need
    # to be re-written/
    try:
        from schema_enforcer.ansible_inventory import AnsibleInventory  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError:
        error(
            "ansible package not found, you can run the command 'pip install schema-enforcer[ansible]' to install the latest schema-enforcer sanctioned version."
        )
        sys.exit(1)

    if inventory:
        config.load(config_data={"ansible_inventory": inventory})
    else:
        config.load()

    # ---------------------------------------------------------------------
    # Load Schema(s) from disk
    # ---------------------------------------------------------------------
    try:
        smgr = SchemaManager(config=config.SETTINGS)
    except InvalidJSONSchema as exc:
        error(str(exc))
        sys.exit(1)

    if not smgr.schemas:
        error("No schemas were loaded")
        sys.exit(1)

    # ---------------------------------------------------------------------
    # Load Ansible Inventory file
    #  - generate hostvar for all devices in the inventory
    #  - Validate Each key in the hostvar individually against the schemas defined in the var jsonschema_mapping
    # ---------------------------------------------------------------------
    inv = AnsibleInventory(inventory=config.SETTINGS.ansible_inventory)
    hosts = inv.get_hosts_containing()
    print(f"Found {len(hosts)} hosts in the inventory")

    if show_checks:
        inv.print_schema_mapping(hosts, limit, smgr)
        sys.exit(0)

    error_exists = False

    for host in hosts:
        if limit and host.name != limit:
            continue

        # Acquire Host Variables
        hostvars = inv.get_clean_host_vars(host)

        # Acquire validation settings for the given host
        schema_validation_settings = inv.get_schema_validation_settings(host)
        declared_schema_ids = schema_validation_settings["declared_schema_ids"]
        strict = schema_validation_settings["strict"]
        automap = schema_validation_settings["automap"]

        # Validate declared schemas exist
        smgr.validate_schemas_exist(declared_schema_ids)

        # Acquire schemas applicable to the given host
        applicable_schemas = inv.get_applicable_schemas(hostvars, smgr, declared_schema_ids, automap)
        for schema_obj in applicable_schemas.values():
            # Combine host attributes into a single data structure matching to properties defined at the top level of the schema definition
            if not strict:
                data = {}
                for var in schema_obj.top_level_properties:
                    data.update({var: hostvars.get(var)})

            # If the schema_enforcer_strict bool is set, hostvars should match a single schema exactly.
            # Thus, we want to pass the entirety of the cleaned host vars into the validate method rather
            # than creating a data structure with only the top level vars defined by the schema.
            else:
                data = hostvars

            # Validate host vars against schema
            schema_obj.validate(data=data, strict=strict)

            for result in schema_obj.get_results():
                result.instance_type = "HOST"
                result.instance_hostname = host.name

                if not result.passed():
                    error_exists = True
                    result.print()

                elif result.passed() and show_pass:
                    result.print()
            schema_obj.clear_results()

    if not error_exists:
        print(colored("ALL SCHEMA VALIDATION CHECKS PASSED", "green"))
    else:
        sys.exit(1)



"""Ansible Inventory class to generate final hostvars based on group_vars and host_vars."""
from ansible.inventory.manager import InventoryManager  # pylint: disable=import-error
from ansible.parsing.dataloader import DataLoader  # pylint: disable=import-error
from ansible.vars.manager import VariableManager  # pylint: disable=import-error
from ansible.template import Templar  # pylint: disable=import-error


# Referenced https://github.com/fgiorgetti/qpid-dispatch-tests/ for the below class
class AnsibleInventory:
    """AnsibleInventory."""

    def __init__(self, inventory=None, extra_vars=None):
        """Imitates Ansible Inventory Loader.

        Args:
            inventory (str): Path to Ansible Inventory files.
            extra_vars (dict): Extra Vars passed at run time.
        """
        self.inventory = inventory
        self.loader = DataLoader()
        self.inv_mgr = InventoryManager(loader=self.loader, sources=self.inventory)
        self.var_mgr = VariableManager(loader=self.loader, inventory=self.inv_mgr)
        # TODO As of Ansible==2.8.0 the extra_vars property cannot be set to VariableManager
        #      This needs to be investigated and fixed properly
        self.extra_vars = extra_vars or {}

    def get_hosts_containing(self, var=None):
        """Gets hosts that have a value for ``var``.

        If ``var`` is None, then all hosts in the inventory will be returned.

        Args:
            var (str): The variable to use to restrict hosts.

        Returns:
            list: All ansible.inventory.host.Host objects that define ``var``.
        """
        all_hosts = self.inv_mgr.get_hosts()
        if var is None:
            return all_hosts

        # Only add hosts that define the variable.
        hosts_with_var = [host for host in all_hosts if var in self.var_mgr.get_vars(host=host)]
        return hosts_with_var

    def get_host_vars(self, host):
        """Retrieves Jinja2 rendered variables for ``host``.

        Args:
            host (ansible.inventory.host.Host): The host to retrieve variable data from.

        Returns:
            dict: The variables defined by the ``host`` in Ansible Inventory.
        """
        data = self.var_mgr.get_vars(host=host)
        templar = Templar(variables=data, loader=self.loader)
        return templar.template(data, fail_on_undefined=False)

    def get_clean_host_vars(self, host):
        """Return clean hostvars for a given host, cleaned up of all keys inserted by Templar.

        Args:
            host (ansible.inventory.host.Host): The host to retrieve variable data from.

        Raises:
            TypeError: When "magic_vars_to_evaluate" is declared in an Ansible inventory file and is not of type list,
            a type error is raised

        Returns:
            dict: clean hostvars
        """
        keys_cleanup = [
            "inventory_file",
            "inventory_dir",
            "inventory_hostname",
            "inventory_hostname_short",
            "group_names",
            "ansible_facts",
            "playbook_dir",
            "ansible_playbook_python",
            "groups",
            "omit",
            "ansible_version",
            "ansible_config_file",
            "schema_enforcer_schema_ids",
            "schema_enforcer_strict",
            "schema_enforcer_automap_default",
            "magic_vars_to_evaluate",
        ]

        hostvars = self.get_host_vars(host)

        # Extract magic vars which should be evaluated
        magic_vars_to_evaluate = hostvars.get("magic_vars_to_evaluate", [])
        if not isinstance(magic_vars_to_evaluate, list):
            raise TypeError(f"magic_vars_to_evaluate variable configured for host {host.name} must be of type list")

        keys_cleanup = list(set(keys_cleanup) - set(magic_vars_to_evaluate))

        for key in keys_cleanup:
            if key in hostvars:
                del hostvars[key]

        return hostvars

    @staticmethod
    def get_applicable_schemas(hostvars, smgr, declared_schema_ids, automap):
        """Get applicable schemas.

        Search an explicit mapping to determine the schemas which should be used to validate hostvars
        for a given host.

        If an explicit mapping is not defined, correlate top level keys in the structured data with top
        level properties in the schema to acquire applicable schemas.

        Args:
            hostvars (dict): dictionary of cleaned host vars which will be evaluated against schema
            smgr (schema_enforcer.schemas.manager.SchemaManager): SchemaManager object
            declared_schema_ids (list): A list of declared schema IDs inferred from schema_enforcer_schemas variable
            automap (bool): Whether or not to use the `automap` feature to automatically map top level hostvar keys
                to top level schema definition properties if no schema ids are declared (list of schema ids is empty)

        Returns:
            applicable_schemas (dict): dictionary mapping schema_id to schema obj for all applicable schemas
        """
        applicable_schemas = {}
        for key in hostvars.keys():
            # extract applicable schema ID to JsonSchema objects if schema_ids are declared
            if declared_schema_ids:
                for schema_id in declared_schema_ids:
                    applicable_schemas[schema_id] = smgr.schemas[schema_id]

            # extract applicable schema ID to JsonSchema objects based on host var to top level property mapping.
            elif automap:
                for schema in smgr.schemas.values():
                    if key in schema.top_level_properties:
                        applicable_schemas[schema.id] = schema
                        continue

        return applicable_schemas

    def get_schema_validation_settings(self, host):
        """Parse Ansible Schema Validation Settings from a host object.

        Validate settings or ensure an error is raised in the event an invalid parameter is
        configured in the host file.

        Args:
            host (AnsibleInventory.host): Ansible Inventory Host Object

        Raises:
            TypeError: Raised when one of the schema configuration parameters is of the wrong type
            ValueError: Raised when one of the schema configuration parameters is incorrectly configured

        Returns:
            (dict): Dict of validation settings with keys "declared_schema_ids", "strict", and "automap"
        """
        # Generate host_var and automatically remove all keys inserted by ansible
        hostvars = self.get_host_vars(host)

        # Extract declared_schema_ids from hostvar setting
        declared_schema_ids = []
        if "schema_enforcer_schema_ids" in hostvars:
            if not isinstance(hostvars["schema_enforcer_schema_ids"], list):
                raise TypeError(f"'schema_enforcer_schema_ids' attribute defined for {host.name} must be of type list")
            declared_schema_ids = hostvars["schema_enforcer_schema_ids"]

        # Extract whether to use a strict validator or a loose validator from hostvar setting
        strict = False
        if "schema_enforcer_strict" in hostvars:
            if not isinstance(hostvars["schema_enforcer_strict"], bool):
                raise TypeError(f"'schema_enforcer_strict' attribute defined for {host.name} must be of type bool")
            strict = hostvars["schema_enforcer_strict"]

        automap = True
        if "schema_enforcer_automap_default" in hostvars:
            if not isinstance(hostvars["schema_enforcer_automap_default"], bool):
                raise TypeError(
                    f"'schema_enforcer_automap_default' attribute defined for {host.name} must be of type bool"
                )
            automap = hostvars["schema_enforcer_automap_default"]

        # Raise error if settings are set incorrectly
        if strict and not declared_schema_ids:
            msg = (
                f"The 'schema_enforcer_strict' parameter is set for {host.name} but the 'schema_enforcer_schema_ids' parameter does not declare a schema id. "
                "The 'schema_enforcer_schema_ids' parameter MUST be defined as a list declaring only one schema ID if 'schema_enforcer_strict' is set."
            )
            raise ValueError(msg)

        if strict and declared_schema_ids and len(declared_schema_ids) > 1:
            msg = (
                f"The 'schema_enforcer_strict' parameter is set for {host.name} but the 'schema_enforcer_schema_ids' parameter declares more than one schema id. "
                "The 'schema_enforcer_schema_ids' parameter MUST be defined as a list declaring only one schema ID if 'schema_enforcer_strict' is set."
            )
            raise ValueError(msg)

        return {
            "declared_schema_ids": declared_schema_ids,
            "strict": strict,
            "automap": automap,
        }

    def print_schema_mapping(self, hosts, limit, smgr):
        """Print host to schema IDs mapping.

        Args:
            hosts (list): A list of ansible.inventory.host.Host objects for which the mapping should be printed
            limit (str): The host to which to limit the search
            smgr (schema_enforcer.schemas.manager.SchemaManager): Schema manager which handles schema objects
        """
        print_dict = {}
        for host in hosts:
            if limit and host.name != limit:
                continue

            # Get hostvars
            hostvars = self.get_clean_host_vars(host)

            # Acquire validation settings for the given host
            schema_validation_settings = self.get_schema_validation_settings(host)
            declared_schema_ids = schema_validation_settings["declared_schema_ids"]
            automap = schema_validation_settings["automap"]

            # Validate declared schemas exist
            smgr.validate_schemas_exist(declared_schema_ids)

            # Acquire schemas applicable to the given host
            applicable_schemas = self.get_applicable_schemas(hostvars, smgr, declared_schema_ids, automap)

            # Add an element to the print dict for this host
            print_dict[host.name] = list(applicable_schemas.keys())

        if print_dict:
            print("{:25} Schema ID".format("Ansible Host"))  # pylint: disable=consider-using-f-string
            print("-" * 80)
            print_strings = []
            for hostname, schema_ids in print_dict.items():
                print_strings.append(f"{hostname:25} {schema_ids}")
            print("\n".join(sorted(print_strings)))
